package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.block.BlockBatteryBank;
import ru.dbotthepony.mc.otm.block.BlockMatteryRotatable;
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.menu.BatteryBankMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;
import java.util.Optional;

public class BlockEntityBatteryBank extends BlockEntityMattery {
	// 6 на 2
	public final MatteryContainer battery_container = new MatteryContainer(this::setChanged, 6 * 2) {
		@Override
		public void setChanged(int slot, ItemStack new_state, ItemStack old_state) {
			super.setChanged(slot, new_state, old_state);

			if (level != null) {
				var state = getBlockState();

				for (int i = 0; i < BlockBatteryBank.BATTERY_SLOTS_PROPS.length; i++) {
					state = state.setValue(BlockBatteryBank.BATTERY_SLOTS_PROPS[i], getItem(i).getCapability(CapabilityEnergy.ENERGY).isPresent());
				}

				if (state != getBlockState()) {
					level.setBlock(getBlockPos(), state, Block.UPDATE_CLIENTS);
				}
			}
		}
	};

	private final LazyOptional<IItemHandler> item_handler_resolver = LazyOptional.of(() -> battery_container.handler(
		(slot, stack) -> stack.getCapability(CapabilityEnergy.ENERGY).isPresent(),
		(slot, amount, stack) -> true
	));

	public record BatteryBankDistribution(BigDecimal[] distribution, BigDecimal max_throughput) {

	}

	public static class BatteryBankEnergy implements IMatteryEnergyStorage {
		public enum BankMode {
			RECEIVE,
			EXTRACT,
			BIDIRECTIONAL
		}

		private final BlockEntityBatteryBank tile;
		private final BankMode mode;

		BatteryBankEnergy(BlockEntityBatteryBank tile, BankMode mode) {
			this.tile = tile;
			this.mode = mode;
		}

		@Override
		public boolean canExtract() {
			return mode == BankMode.EXTRACT || mode == BankMode.BIDIRECTIONAL;
		}

		@Override
		public boolean canReceive() {
			return mode == BankMode.RECEIVE || mode == BankMode.BIDIRECTIONAL;
		}

		@Nonnull
		@Override
		public BigDecimal extractEnergyOuter(BigDecimal howMuch, boolean simulate) {
			if (mode == BankMode.RECEIVE)
				return BigDecimal.ZERO;

			return extractEnergyInner(howMuch, simulate);
		}

		BatteryBankDistribution getDistribution(boolean mode) {
			BigDecimal[] distribution = new BigDecimal[tile.battery_container.getContainerSize()];
			BigDecimal summ = BigDecimal.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				ItemStack stack = tile.battery_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

					if (cap.isPresent()) {
						BigDecimal diff = mode ? cap.get().receiveEnergyOuter(MatteryCapability.LONG_MAX_VALUE, true) : cap.get().extractEnergyOuter(MatteryCapability.LONG_MAX_VALUE, true);
						distribution[i] = diff;
						summ = summ.add(diff);
					} else {
						Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

						if (cap2.isPresent()) {
							BigDecimal diff = new BigDecimal(mode ? cap2.get().receiveEnergy(Integer.MAX_VALUE, true) : cap2.get().extractEnergy(Integer.MAX_VALUE, true));
							distribution[i] = diff;
							summ = summ.add(diff);
						}
					}
				} else {
					distribution[i] = BigDecimal.ZERO;
				}
			}

			if (summ.compareTo(BigDecimal.ZERO) != 0) {
				for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
					distribution[i] = distribution[i].divide(summ, MatteryCapability.ROUND_RULES);
				}
			}

			return new BatteryBankDistribution(distribution, summ);
		}

		private BigDecimal distributeEnergy(boolean mode, BigDecimal howMuch, boolean simulate) {
			BatteryBankDistribution distribution = getDistribution(mode);

			if (distribution.max_throughput.compareTo(BigDecimal.ZERO) == 0)
				return BigDecimal.ZERO;

			BigDecimal[] distList = distribution.distribution;
			BigDecimal summ = BigDecimal.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				if (!distList[i].equals(BigDecimal.ZERO)) {
					ItemStack stack = tile.battery_container.getItem(i);

					if (!stack.isEmpty()) {
						Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

						if (cap.isPresent()) {
							BigDecimal diff = mode ? cap.get().receiveEnergyOuter(howMuch.multiply(distList[i], MatteryCapability.ROUND_RULES), simulate) : cap.get().extractEnergyOuter(howMuch.multiply(distList[i], MatteryCapability.ROUND_RULES), simulate);
							summ = summ.add(diff);
						} else {
							Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

							if (cap2.isPresent()) {
								BigDecimal diff = mode ? MatteryCapability.floodFE(cap2.get(), howMuch.multiply(distList[i], MatteryCapability.ROUND_RULES), simulate) : MatteryCapability.drainFE(cap2.get(), howMuch.multiply(distList[i], MatteryCapability.ROUND_RULES), simulate);
								summ = summ.add(diff);
							}
						}
					}
				}
			}

			return summ;
		}

		@Nonnull
		@Override
		public BigDecimal extractEnergyInner(BigDecimal howMuch, boolean simulate) {
			return distributeEnergy(false, howMuch, simulate);
		}

		@Nonnull
		@Override
		public BigDecimal receiveEnergyOuter(BigDecimal howMuch, boolean simulate) {
			if (mode == BankMode.EXTRACT)
				return BigDecimal.ZERO;

			return receiveEnergyInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public BigDecimal receiveEnergyInner(BigDecimal howMuch, boolean simulate) {
			return distributeEnergy(true, howMuch, simulate);
		}

		@Nonnull
		@Override
		public BigDecimal getBatteryLevel() {
			BigDecimal result = BigDecimal.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				ItemStack stack = tile.battery_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

					if (cap.isPresent()) {
						result = result.add(cap.get().getBatteryLevel(), MatteryCapability.ROUND_RULES);
					} else {
						Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

						if (cap2.isPresent()) {
							result = result.add(new BigDecimal(cap2.get().getEnergyStored()), MatteryCapability.ROUND_RULES);
						}
					}
				}
			}

			return result;
		}

		@Nonnull
		@Override
		public BigDecimal getMaxBatteryLevel() {
			BigDecimal result = BigDecimal.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				ItemStack stack = tile.battery_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

					if (cap.isPresent()) {
						result = result.add(cap.get().getMaxBatteryLevel(), MatteryCapability.ROUND_RULES);
					} else {
						Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

						if (cap2.isPresent()) {
							result = result.add(new BigDecimal(cap2.get().getMaxEnergyStored()), MatteryCapability.ROUND_RULES);
						}
					}
				}
			}

			return result;
		}
	}

	public final BatteryBankEnergy energy_receiver = new BatteryBankEnergy(this, BatteryBankEnergy.BankMode.RECEIVE);
	public final LazyOptional<BatteryBankEnergy> energy_receiver_resolver = LazyOptional.of(() -> energy_receiver);

	public final BatteryBankEnergy energy_extractor = new BatteryBankEnergy(this, BatteryBankEnergy.BankMode.EXTRACT);
	public final LazyOptional<BatteryBankEnergy> energy_extractor_resolver = LazyOptional.of(() -> energy_extractor);

	public final BatteryBankEnergy energy = new BatteryBankEnergy(this, BatteryBankEnergy.BankMode.BIDIRECTIONAL);
	public final LazyOptional<BatteryBankEnergy> energy_resolver = LazyOptional.of(() -> energy);

	protected boolean valid = true;

	public BlockEntityBatteryBank(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.BATTERY_BANK, p_155229_, p_155230_);
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.put("battery_bank", battery_container.serializeNBT());
		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		battery_container.deserializeNBT(nbt.get("battery_bank"));
		super.load(nbt);
	}

	private final TranslatableComponent MACHINE_NAME = new TranslatableComponent("block.overdrive_that_matters.battery_bank");

	@Override
	protected Component getDefaultDisplayName() {
		return MACHINE_NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new BatteryBankMenu(containerID, inventory, this, this.battery_container);
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
	}

	@Nonnull
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid && (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY)) {
			if (side == null)
				return energy_resolver.cast();

			if (side == getBlockState().getValue(BlockMatteryRotatable.FACING))
				return energy_extractor_resolver.cast();

			return energy_receiver_resolver.cast();
		}

		if (valid && cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
			return item_handler_resolver.cast();

		return super.getCapability(cap, side);
	}

	public static <T extends BlockEntity> void tick(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityBatteryBank tile) {
			BlockEntity get_entity = level.getBlockEntity(blockPos.offset(tile.getBlockState().getValue(BlockMatteryRotatable.FACING).getNormal()));

			if (get_entity != null) {
				Optional<IMatteryEnergyStorage> cap = get_entity.getCapability(MatteryCapability.ENERGY).resolve();

				if (cap.isPresent()) {
					if (cap.get().canReceive()) {
						BatteryBankDistribution distribution = tile.energy.getDistribution(false);
						BigDecimal diff = cap.get().receiveEnergyOuter(distribution.max_throughput, true);
						diff = tile.energy.extractEnergyOuter(diff, false);
						cap.get().receiveEnergyOuter(diff, false);
					}
				} else {
					Optional<IEnergyStorage> cap2 = get_entity.getCapability(CapabilityEnergy.ENERGY).resolve();

					if (cap2.isPresent()) {
						if (cap2.get().canReceive()) {
							BatteryBankDistribution distribution = tile.energy.getDistribution(false);
							BigDecimal diff = MatteryCapability.floodFE(cap2.get(), distribution.max_throughput, true);
							diff = tile.energy.extractEnergyOuter(diff, false);
							MatteryCapability.floodFE(cap2.get(), diff, false);
						}
					}
				}
			}
		}
	}
}
