package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.*;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;
import ru.dbotthepony.mc.otm.menu.MatterReplicatorMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;

public class BlockEntityMatterReplicator extends BlockEntityMatteryPoweredWorker implements IMatterGridCell {
	public BlockEntityMatterReplicator(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.MATTER_REPLICATOR, p_155229_, p_155230_);
		energy = new MatteryMachineEnergyStorage(this, MatteryMachineEnergyStorage.MachineType.WORKER, new BigDecimal(200_000), new BigDecimal(4000), new BigDecimal(4000));
	}

	public final MatterHandlerCapability matter = new MatterHandlerCapability(this::setChanged, IMatterHandler.MatterDirection.RECEIVE, new BigDecimal(2));

	// обычные запросы
	public final MatteryContainer regular_slots = new MatteryContainer(this::setChanged, 3);

	// запросы от matter replicator interface
	public final MatteryContainer reserved_slots = new MatteryContainer(this::setChanged, 3);

	private final LazyOptional<IItemHandler> resolve_handler = LazyOptional.of(() -> regular_slots.handler(
		(slot, stack) -> false
	));

	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.matter_replicator");

	private static final BigDecimal BASE_CONSUMPTION = new BigDecimal(400);

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new MatterReplicatorMenu(containerID, inventory, this);
	}

	@Nonnull
	@Override
	protected BigDecimal getBaseConsumption() {
		return BASE_CONSUMPTION;
	}

	@Nonnull
	@Override
	protected MachineJobStatus onJobFinish(MachineJob job) {
		if (!regular_slots.addItem(job.stack()).isEmpty()) {
			return new MachineJobStatus(false, 20);
		}

		if (grid != null) {
			grid.notifyTaskCompletion(MatterTask.deserializeNBT(job.data().get("task")));
		}

		return new MachineJobStatus();
	}

	@Override
	public void onMatterTaskCreated(MatterTask task) {
		is_idling = false;
	}

	@Override
	public void onMatterTaskUpdated(MatterTask new_state, MatterTask old_state) {
		is_idling = false;
	}

	@Override
	public void onPatternAdded(PatternState state) {
		is_idling = false;
	}

	@Override
	public void setRemoved() {
		if (level != null && !level.isClientSide && grid != null)
			grid.untrack(this);

		super.setRemoved();
	}

	private static final double TICKS_PER_MTU = 20_000d;
	private static final BigDecimal TICKS_PER_MTU_BD = new BigDecimal(20_000);
	private static final double MTU_PER_TICK = 1d / 20_000d;
	private static final BigDecimal MTU_PER_TICK_BD = BigDecimal.ONE.divide(TICKS_PER_MTU_BD, MatteryCapability.ROUND_RULES);

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			scheduleDiscoverNeighbours(getBlockPos(), p_155231_);
	}

	@Nullable
	@Override
	protected MachineJob getNextJob() {
		if (grid == null)
			return null;

		MatterTaskAllocation allocation = grid.allocateTask(false);

		if (allocation == null)
			return null;

		ItemStack stack = allocation.task().stack(1);

		// ????????
		if (!MatterRegistry.hasMatterValue(stack))
			return null;

		MachineJob job = new MachineJob(stack, MatterRegistry.getMatterValue(stack).doubleValue() * TICKS_PER_MTU);

		job.data().put("task", allocation.task().serializeNBT());

		if (allocation.pattern() != null)
			job.data().put("pattern", allocation.pattern().serializeNBT());

		return job;
	}

	private static final BigDecimal DRAIN_MULT = new BigDecimal(200);

	@Override
	protected MachineJobStatus onWorkTick(WorkTickContext context) {
		BigDecimal drain_per_tick = MTU_PER_TICK_BD.multiply(context.work_speed(), MatteryCapability.ROUND_RULES);

		if (matter.extractMatterInner(drain_per_tick, true).compareTo(drain_per_tick) < 0) {
			// в машине недостаточно материи
			if (grid == null)
				return new MachineJobStatus(false, 20);

			if (drain_per_tick.compareTo(matter.getMaxStoredMatter()) > 0) {
				// в тик требуется больше материи, чем её может хранить репликатор
				BigDecimal to_extract = drain_per_tick.subtract(matter.extractMatterInner(drain_per_tick, true));
				BigDecimal drain = grid.extractMatter(to_extract, true);

				if (drain.compareTo(to_extract) < 0) {
					// недостаточно материи в сети
					return new MachineJobStatus(false, 200);
				}

				// достаточно материи в сети + внутри машины
				matter.extractMatterInner(drain_per_tick, false);
				grid.extractMatter(drain, false);

				return new MachineJobStatus();
			} else {
				// в тик требуется меньше материи, чем её может хранить репликатор
				// примем из сети недостающее количество бака материи, или 200 тиков репликации, что меньше
				BigDecimal to_extract = matter.getMissingMatter().min(drain_per_tick.multiply(DRAIN_MULT));
				BigDecimal drain = grid.extractMatter(to_extract, true);

				if (drain.compareTo(BigDecimal.ZERO) == 0) {
					// в сети нет материи
					return new MachineJobStatus(false, 200);
				}

				BigDecimal received = matter.receiveMatterOuter(drain, false);
				grid.extractMatter(received, false);

				// получили материю, проверяем возможность работы
				if (matter.extractMatterInner(drain_per_tick, false).compareTo(drain_per_tick) >= 0) {
					return new MachineJobStatus();
				} else {
					// :(
					return new MachineJobStatus(false, 200);
				}
			}
		}

		// в машине достаточно материи

		matter.extractMatterInner(drain_per_tick, false);
		return new MachineJobStatus();
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.put("regular_slots", regular_slots.serializeNBT());
		nbt.put("reserved_slots", reserved_slots.serializeNBT());
		nbt.put("matter_capability", matter.serializeNBT());
		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);
		regular_slots.deserializeNBT(nbt.get("regular_slots"));
		reserved_slots.deserializeNBT(nbt.get("reserved_slots"));

		if (nbt.get("matter_capability") instanceof CompoundTag tag)
			matter.deserializeNBT(tag);
	}

	private boolean valid = true;

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
	}

	private final LazyOptional<BlockEntityMatterReplicator> resolver_self = LazyOptional.of(() -> this);

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid) {
			if (cap == MatteryCapability.MATTER_CELL)
				return resolver_self.cast();

			if (cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
				return resolve_handler.cast();
		}

		return super.getCapability(cap, side);
	}

	private MatterGrid grid;

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}
}
