package ru.dbotthepony.mc.otm.menu;

import net.minecraft.core.BlockPos;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.inventory.ContainerData;
import net.minecraft.world.inventory.MenuType;
import net.minecraft.world.inventory.Slot;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.block.entity.BlockEntity;
import ru.dbotthepony.mc.otm.menu.slot.MatterySlot;
import ru.dbotthepony.mc.otm.menu.widget.AbstractWidget;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.function.Consumer;

public abstract class MatteryMenu extends AbstractContainerMenu {
	protected BlockEntity tile;

	protected Inventory inventory;
	protected Player ply;

	public final ArrayList<AbstractWidget> mattery_widgets = new ArrayList<>();
	public final ArrayList<MatterySlot> inventory_slots = new ArrayList<>();
	public final ArrayList<MatterySlot> main_slots = new ArrayList<>();

	protected MatteryMenu(@Nullable MenuType<?> p_38851_, int p_38852_, Inventory inventory) {
		super(p_38851_, p_38852_);
		this.inventory = inventory;
		this.ply = inventory.player;
	}

	protected MatteryMenu(@Nullable MenuType<?> p_38851_, int p_38852_, Inventory inventory, BlockEntity tile) {
		this(p_38851_, p_38852_, inventory);
		this.tile = tile;
	}

	public void addWidget(AbstractWidget widget, Consumer<Consumer<ContainerData>> consumer) {
		if (mattery_widgets.contains(widget))
			return;

		mattery_widgets.add(widget);
		consumer.accept(this::addDataSlots);
	}

	public MatterySlot addMainSlot(MatterySlot slot) {
		addSlot(slot);
		main_slots.add(slot);
		return slot;
	}

	protected void addInventorySlots() {
		addInventorySlots(97);
	}

	protected int inventory_slot_index_start;
	protected int inventory_slot_index_end;

	protected void addInventorySlots(int offset) {
		boolean first = true;

		for (int i = 0; i < 3; ++i) {
			for (int j = 0; j < 9; ++j) {
				var slot = new MatterySlot(inventory, j + i * 9 + 9, 8 + j * 18, 14 + i * 18);
				inventory_slots.add(slot);
				this.addSlot(slot);

				if (first) {
					first = false;
					inventory_slot_index_start = slot.index;
				}
			}
		}

		MatterySlot last = null;

		for (int k = 0; k < 9; ++k) {
			last = new MatterySlot(inventory, k, 8 + k * 18, 14 + 58);
			this.addSlot(last);
			inventory_slots.add(last);
		}

		inventory_slot_index_end = last.index;
	}

	@Override
	public void broadcastChanges() {
		for (AbstractWidget widget : this.mattery_widgets) {
			widget.updateServer();
		}

		super.broadcastChanges();
	}

	@Override
	public void broadcastFullState() {
		for (AbstractWidget widget : this.mattery_widgets) {
			widget.updateServer();
		}

		super.broadcastFullState();
	}

	@Override
	public boolean stillValid(Player player) {
		if (tile == null)
			return true;

		if (player.level.getBlockEntity(tile.getBlockPos()) != tile) {
			return false;
		}

		BlockPos pos = tile.getBlockPos();
		return player.distanceToSqr((double)pos.getX() + 0.5D, (double)pos.getY() + 0.5D, (double)pos.getZ() + 0.5D) <= 64.0D;
	}

	abstract protected int getWorkingSlotStart();
	abstract protected int getWorkingSlotEnd();

	protected void notifySlotsWatchCurrentStack() {
		for (var slot : slots) {
			if (slot instanceof MatterySlot slot1) {
				slot1.watchCurrentStack();
			}
		}
	}

	protected void notifySlotsUnwatchCurrentStack() {
		for (var slot : slots) {
			if (slot instanceof MatterySlot slot1) {
				slot1.unwatchCurrentStack();
			}
		}
	}

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	@Override
	public ItemStack quickMoveStack(Player ply, int slot_index) {
		// this.moveItemStackTo(ItemStack, int start_slot_index, int end_slot_index, boolean iteration_order)
		// returns boolean, telling whenever ItemStack was modified (moved or sharnk)
		// false means nothing happened
		// Last boolean determine order of slot iteration, where:
		// if TRUE, iteration order is end_slot_index -> start_slot_index
		// if FALSE iteration order is start_slot_index -> end_slot_index

		int start = getWorkingSlotStart();
		int end = getWorkingSlotEnd();

		if (start == end) {
			return ItemStack.EMPTY;
		}

		ItemStack moved = ItemStack.EMPTY;
		Slot get_slot = this.slots.get(slot_index);
		notifySlotsWatchCurrentStack();

		if (get_slot.hasItem()) {
			ItemStack slot_item = get_slot.getItem();
			moved = slot_item.copy();

			if (slot_index < inventory_slot_index_start) {
				// Moving FROM machine TO inventory

				if (!moveItemStackTo(slot_item, inventory_slot_index_start, inventory_slot_index_end + 1, false)) {
					notifySlotsUnwatchCurrentStack();
					return ItemStack.EMPTY;
				}
			} else if (!moveItemStackTo(slot_item, start, end, false)) {
				// Moving FROM inventory TO machine
				notifySlotsUnwatchCurrentStack();
				return ItemStack.EMPTY;
			}

			if (slot_item.isEmpty()) {
				get_slot.set(ItemStack.EMPTY);
			} else {
				get_slot.setChanged();
			}

			notifySlotsWatchCurrentStack();
		}

		notifySlotsUnwatchCurrentStack();
		return moved;
	}
}
